import stare
import time
import pickle
import logging
import pytest

# set some things on user that we expect to persist or not persist
_name = 'Test User'
_response = 'A response'
_status_code = 200
_access_token = '4cce$$T0k3n'
_raw_id_token = 'R4w1DT0k3n'


@pytest.fixture
def user_temp(tmpdir):
    temp = tmpdir.join("auth.pkl")
    assert temp.isfile() == False

    u = stare.core.User(save_auth=temp.strpath)
    u._id_token = {'exp': time.time() + 3600, 'userLogin': _name}
    u._response = _response
    u._status_code = _status_code
    u._access_token = _access_token
    u._raw_id_token = _raw_id_token
    return u, temp


def test_user_expires(user_temp):
    user, _ = user_temp
    assert user.is_authenticated()
    assert user.is_expired() == False
    assert user.expires_in > 0
    user._id_token['exp'] = time.time() - 1
    assert user.is_authenticated()
    assert user.is_expired()
    assert user.expires_in == 0


def test_user_serialization(user_temp, caplog):
    user, temp = user_temp
    # set up first user and check that we can dump
    session = stare.core.Session(user=user)
    assert temp.isfile() == False
    assert session.user._dump()
    assert temp.isfile()
    assert temp.size()
    assert pickle.load(open(user._save_auth, 'rb'))

    # check if we can reload user
    session.user._id_token = None
    assert session.user._load()
    assert session.user.name == _name
    del session

    # check if session can load user
    session = stare.core.Session(user=user, save_auth=temp.strpath)
    assert session.user.name == _name
    assert session.user._session
    assert session.user._response == _response
    assert session.user._status_code == _status_code
    assert session.user._access_token == _access_token
    assert session.user._raw_id_token == _raw_id_token

    # check what happens if the saved user has expired
    session.user._id_token['exp'] = time.time() + 4
    session.user._dump()
    time.sleep(5)
    with caplog.at_level(logging.INFO, 'stare'):
        assert session.user._load() == False
        assert 'Expired saved user session in' in caplog.text
    del session

    # check what happens if corruption
    temp.write('fake')
    with caplog.at_level(logging.INFO, 'stare'):
        user = stare.core.User(save_auth=temp.strpath)
        assert 'Unable to load user session' in caplog.text
        caplog.clear()
        assert user._load() == False
        assert 'Unable to load user session' in caplog.text
        caplog.clear()
        stare.core.Session(save_auth=temp.strpath)
        assert 'Unable to load user session' in caplog.text



import stare
import pytest
import requests


@pytest.fixture
def session(mocker):
    mocker.patch('stare.core.User.authenticate', return_value=True)
    mocker.patch('stare.core.Session._handle_response', return_value=True)
    # patch objects
    s = stare.core.Session()
    mocker.patch.object(
        requests.Session, 'request', wraps=super(stare.core.Session, s).request
    )
    mocker.patch.object(stare.core.Session, 'request', wraps=s.request)

    mocker.patch('requests.Session.send', return_value=None)
    mocker.patch.object(stare.core.Session, 'send', wraps=s.send)

    mocker.patch.object(
        requests.Session,
        'prepare_request',
        wraps=super(stare.core.Session, s).prepare_request,
    )
    mocker.patch.object(stare.core.Session, 'prepare_request', wraps=s.prepare_request)

    mocker.patch.object(stare.core.Session, '_normalize_url', wraps=s._normalize_url)
    return s


# FIXME
def test_session_handle_response():
    session = stare.core.Session()
    pass


def test_session_normalize_url():
    session = stare.core.Session()
    url = 'analyses'
    new_url = session._normalize_url(url)
    assert url != new_url
    assert new_url.startswith(stare.settings.SITE_URL)


# NB: Session.request is delegated through to Session.send
def test_session_request(session):
    session.get('analyses')
    assert session._normalize_url.called
    assert session.request.called
    assert super(stare.core.Session, session).request.called
    assert session.send.called
    assert super(stare.core.Session, session).send.called
    assert session._handle_response.call_count == 1


def test_session_request_call(session):
    session('GET', 'analyses')
    assert session._normalize_url.called
    assert session.request.called
    assert super(stare.core.Session, session).request.called
    assert session.send.called
    assert super(stare.core.Session, session).send.called
    assert session._handle_response.call_count == 1


def test_session_prepare_request(session):
    req = requests.Request('GET', 'analyses')
    prep = session.prepare_request(req)
    assert session._normalize_url.called
    assert 'Authorization' in prep.headers
    assert super(stare.core.Session, session).prepare_request.called
    assert session._handle_response.call_count == 0


def test_session_send(session):
    req = requests.Request('GET', 'analyses')
    prep = session.prepare_request(req)
    session.send(prep)
    assert super(stare.core.Session, session).send.called
    assert session._handle_response.call_count == 1


def test_session_send_call(session):
    req = requests.Request('GET', 'analyses')
    session(req)
    assert super(stare.core.Session, session).send.called
    assert session._handle_response.call_count == 1



import socket
from sys import platform
import stare

import betamax
from betamax_serializers import pretty_json

import pytest

placeholders = {'apiKey': stare.settings.GLANCE_API_KEY}


def filter_bearer_token(interaction, current_cassette):
    # Exit early if the request did not return 200 OK because that's the
    # only time we want to look for Authorization-Token headers
    if interaction.data['response']['status']['code'] != 200:
        return

    headers = interaction.data['request']['headers']
    token = headers.get('Authorization')
    # If there was no token header in the request, exit
    if token is None:
        return

    # Otherwise, create a new placeholder so that when cassette is saved,
    # Betamax will replace the token with our placeholder.
    current_cassette.placeholders.append(
        betamax.cassette.cassette.Placeholder(
            placeholder='Bearer <ACCESS_TOKEN>', replace=token[0]
        )
    )


betamax.Betamax.register_serializer(pretty_json.PrettyJSONSerializer)
with betamax.Betamax.configure() as config:
    config.cassette_library_dir = stare.settings.CASSETTE_LIBRARY_DIR
    config.default_cassette_options['serialize_with'] = 'prettyjson'
    config.before_record(callback=filter_bearer_token)
    for key, value in placeholders.items():
        config.define_cassette_placeholder('<{}>'.format(key.upper()), replace=value)


@pytest.fixture(scope='session')
def auth_user():
    user = stare.core.User()
    user._jwtOptions = {
        'verify_signature': False,
        'verify_iat': False,
        'verify_exp': False,
    }
    with betamax.Betamax(
        user._session, cassette_library_dir=stare.settings.CASSETTE_LIBRARY_DIR
    ).use_cassette('test_user.test_user_good_login', record='none') as recorder:
        yield user


@pytest.fixture(scope='module')
def auth_session(auth_user):
    yield stare.core.Session(user=auth_user)


# Temporarily work around issue with gethostbyname on OS X
#  - see https://betamax.readthedocs.io/en/latest/long_term_usage.html#known-issues
if platform == 'darwin':
    socket.gethostbyname = lambda x: '127.0.0.1'



import logging
import requests
from requests.status_codes import codes
from cachecontrol import CacheControlAdapter
from cachecontrol.caches.file_cache import FileCache

from jose import jwt
import time
import os
import pickle  # nosec
import copy

from .settings import settings
from . import exceptions

log = logging.getLogger(__name__)


class User(object):
    def __init__(
        self,
        apiKey=settings.GLANCE_API_KEY,
        audience='#FIXME',
        jwtOptions={'verify_signature': False},  # FIXME
        save_auth=None,
        verify=settings.CERN_SSL_CHAIN,
    ):

        # session handling (for injection in tests)
        self._session = requests.Session()
        self._session.verify = verify
        # store last call to authenticate
        self._response = None
        self._status_code = None
        # store jwks for validation/verification
        self._jwks = None
        # store information after authorization occurs
        self._access_token = None
        self._raw_id_token = None
        self._id_token = None
        # initialization configuration
        self._apiKey = apiKey
        self._audience = audience
        self._jwtOptions = jwtOptions
        # serialization/persistence
        self._save_auth = save_auth
        self._load()

    def _load(self):
        if self._save_auth and os.path.isfile(self._save_auth):
            try:
                saved_user = pickle.load(open(self._save_auth, 'rb'))  # nosec
                if saved_user.is_expired():
                    log.warning(
                        "Expired saved user session in {}. Creating a new one.".format(
                            self._save_auth
                        )
                    )
                    return False
                if saved_user.is_authenticated():
                    self.__dict__.update(saved_user.__dict__)
                    return True
                return False
            except pickle.UnpicklingError:
                log.warning(
                    "Unable to load user session from {}. Creating a new one.".format(
                        self._save_auth
                    )
                )
            except KeyError:  # python2 specific error
                log.warning(
                    "Unable to load user session from {}. Creating a new one.".format(
                        self._save_auth
                    )
                )
            return False

    def _dump(self):
        if self.is_authenticated() and not self.is_expired() and self._save_auth:
            try:
                pickle.dump(self, open(self._save_auth, 'wb'), pickle.HIGHEST_PROTOCOL)
                return True
            except pickle.PicklingError:
                log.warning(
                    "Unable to save user session to {}.".format(self._save_auth)
                )
            return False

    def _load_jwks(self, force=False):
        if self._jwks is None or force:
            self._jwks = self._session.get('#FIXME').json()

    def _parse_id_token(self):
        if self._id_token:
            # self._load_jwks()  # FIXME
            self._id_token = jwt.decode(
                self._id_token,
                self._jwks,  # FIXME
                algorithms='RS256',
                audience=self._audience,
                options=self._jwtOptions,
            )

    def authenticate(self):
        # if not expired, do nothing
        if self.is_authenticated():
            return True
        # session-less request
        response = self._session.get(
            requests.compat.urljoin(settings.SITE_URL, 'auth'),
            headers={'Api-Key': self._apiKey},
        )
        self._response = response.json()
        self._status_code = response.status_code
        self._access_token = self._response.get('accessToken')
        self._raw_id_token = self._response.get('accessToken')
        self._id_token = self._raw_id_token

        # handle parsing the id token
        self._parse_id_token()

        if not self.is_authenticated():
            log.warning(
                'Authorization failed. Message: {}'.format(self._response['message'])
            )
        else:
            self._dump()

    @property
    def apiKey(self):
        return self._apiKey

    @property
    def access_token(self):
        return self._access_token

    @property
    def id_token(self):
        return self._id_token if self._id_token else {}

    @property
    def id(self):
        return self.id_token.get('userId', '')

    @property
    def name(self):
        return self.id_token.get('userLogin', '')

    @property
    def permissions(self):
        return self.id_token.get('permissions', [])

    @property
    def egroups(self):
        return self.id_token.get('egroups', [])

    @property
    def usergroups(self):
        return self.id_token.get('usergroups', [])

    @property
    def expires_at(self):
        return self.id_token.get('exp', 0)

    @property
    def expires_in(self):
        expires_in = self.expires_at - time.time()
        return 0 if expires_in < 0 else int(expires_in)

    @property
    def bearer(self):
        return self._raw_id_token if self._raw_id_token else ''

    def is_authenticated(self):
        return bool(
            self._status_code == codes['ok']
            and self._access_token
            and self._raw_id_token
        )

    def is_expired(self):
        return not (self.expires_in > 0)

    def __repr__(self):
        return "{0:s}(name={1:s}, expires_in={2:d}s)".format(
            self.__class__.__name__, self.name, self.expires_in
        )


class Session(requests.Session):
    STATUS_EXCEPTIONS = {
        codes['bad_gateway']: exceptions.ServerError,
        codes['bad_request']: exceptions.BadRequest,
        codes['conflict']: exceptions.Conflict,
        codes['found']: exceptions.Redirect,
        codes['forbidden']: exceptions.Forbidden,
        codes['gateway_timeout']: exceptions.ServerError,
        codes['internal_server_error']: exceptions.ServerError,
        codes['media_type']: exceptions.SpecialError,
        codes['not_found']: exceptions.NotFound,
        codes['request_entity_too_large']: exceptions.TooLarge,
        codes['service_unavailable']: exceptions.ServerError,
        codes['unauthorized']: exceptions.Forbidden,
        codes['unavailable_for_legal_reasons']: exceptions.UnavailableForLegalReasons,
    }
    SUCCESS_STATUSES = {codes['created'], codes['ok']}

    def __init__(
        self,
        user=None,
        prefix_url=settings.SITE_URL,
        save_auth=None,
        verify=settings.CERN_SSL_CHAIN,
    ):
        super(Session, self).__init__()
        self.user = user if user else User(save_auth=save_auth)
        self.auth = self._authorize
        self.prefix_url = prefix_url
        self.verify = verify
        # store last call
        self._response = None
        # add caching
        super(Session, self).mount(
            self.prefix_url, CacheControlAdapter(cache=FileCache('.webcache'))
        )

    def _authorize(self, req):
        self.user.authenticate()
        req.headers.update({'Authorization': 'Bearer {0:s}'.format(self.user.bearer)})
        return req

    def _normalize_url(self, url):
        if self.prefix_url not in url:
            return requests.compat.urljoin(self.prefix_url, url)
        return url

    def _handle_response(self, response):
        self._response = response
        log.debug(
            'Response: {} ({} bytes)'.format(
                response.status_code, response.headers.get('content-length')
            )
        )
        if response.status_code in self.STATUS_EXCEPTIONS:
            raise self.STATUS_EXCEPTIONS[response.status_code](response)

        if response.status_code in self.SUCCESS_STATUSES:
            if response.headers.get('content-length') == '0':
                return ''
            try:
                return response.json()
            except ValueError:
                raise exceptions.BadJSON(response)
        else:
            raise exceptions.UnhandledResponse(response)

    def prepare_request(self, request):
        request = copy.deepcopy(request)
        request.url = self._normalize_url(request.url)
        return super(Session, self).prepare_request(request)

    def send(self, request, **kwargs):
        response = super(Session, self).send(request, **kwargs)
        return self._handle_response(response)

    def request(self, method, url, *args, **kwargs):
        url = self._normalize_url(url)
        return super(Session, self).request(method, url, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        if len(args) == 1:
            return self.send(self.prepare_request(*args), **kwargs)
        else:
            return self.request(*args, **kwargs)



import stare
import betamax


def test_get(auth_session):
    with betamax.Betamax(auth_session).use_cassette('test_analysis.test_get'):
        response = auth_session.get('analyses')
        assert auth_session._response.status_code == 200
        assert response
        assert 'analyses' in response
        assert stare.models.analysis.make_analysis_list(response['analyses'])



from .version import __version__
from .core import Session
from .settings import settings
from requests import Request


class Glance(object):
    def __init__(self, session=None):
        self.session = session if session else Session()

    @property
    def analyses(self):
        return self.session(Request('GET', 'analyses'))

    @property
    def papers(self):
        return self.session(Request('GET', 'papers'))


__all__ = ['__version__', 'Session', 'settings', 'Glance']



import pytest
from click.testing import CliRunner

import time
import betamax

import stare
from stare import commandline


@pytest.fixture(scope='module')
def recorder_session(auth_user):
    commandline.client.session.user = auth_user
    with betamax.Betamax(
        commandline.client.session,
        cassette_library_dir=stare.settings.CASSETTE_LIBRARY_DIR,
    ) as recorder:
        yield recorder


def test_commandline():
    assert commandline.client.session
    assert commandline.client.session.user


def test_version():
    runner = CliRunner()
    start = time.time()
    result = runner.invoke(commandline.stare, ['--version'])
    end = time.time()
    elapsed = end - start
    assert result.exit_code == 0
    assert stare.__version__ in result.stdout
    # make sure it took less than a second
    assert elapsed < 1.0


def test_authenticate(recorder_session):
    runner = CliRunner()
    result = runner.invoke(commandline.stare, ['authenticate'])
    assert result.exit_code == 0
    assert 'You have signed in as' in result.output


def test_list_analyses(recorder_session):
    recorder_session.use_cassette('test_analysis.test_get', record='none')
    runner = CliRunner()
    result = runner.invoke(commandline.stare, ['list-analyses'])
    assert result.exit_code == 0
    assert result.output


def test_list_papers(recorder_session):
    recorder_session.use_cassette('test_paper.test_get', record='none')
    runner = CliRunner()
    result = runner.invoke(commandline.stare, ['list-papers'])
    assert result.exit_code == 0
    assert result.output



from dotenv import load_dotenv
import pkg_resources

load_dotenv()

import os

try:
    from types import SimpleNamespace
except:
    from argparse import Namespace as SimpleNamespace

settings = SimpleNamespace(
    GLANCE_API_KEY=os.getenv('GLANCE_API_KEY', ''),
    SITE_URL=os.getenv(
        'SITE_URL',
        'https://glance-stage.cern.ch/release/frapi/atlas/analysis/api/public/',
    ),
    CASSETTE_LIBRARY_DIR=os.getenv(
        'CASSETTE_LIBRARY_DIR', 'tests/integration/cassettes'
    ),
    CERN_SSL_CHAIN=os.getenv(
        'CERN_SSL_CHAIN',
        pkg_resources.resource_filename(
            __name__, os.path.join('data', 'CERN_chain.pem')
        ),
    ),
)



import os
from setuptools import setup, find_packages

extras_require = {
    'develop': [
        'pyflakes',
        'pytest',
        'pytest-cov',
        'pytest-mock',
        'coverage',
        'bumpversion',
        'pre-commit',
        'bandit',
        'black;python_version>="3.6"',  # Black is Python3 only
        'betamax',  # recording api calls for testing
        'betamax-serializers',
        'twine',  # uploading to pypi
    ]
}
extras_require['complete'] = sorted(set(sum(extras_require.values(), [])))

with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md'),
    encoding='utf-8',
) as readme_md:
    long_description = readme_md.read()


def _is_test_pypi():
    """
    Determine if the CI environment has TESTPYPI_UPLOAD defined and
    set to true (c.f. .travis.yml)
    The use_scm_version kwarg accepts a callable for the local_scheme
    configuration parameter with argument "version". This can be replaced
    with a lambda as the desired version structure is {next_version}.dev{distance}
    c.f. https://github.com/pypa/setuptools_scm/#importing-in-setuppy
    As the scm versioning is only desired for TestPyPI, for depolyment to PyPI the version
    controlled through bumpversion is used.
    """
    from os import getenv

    return (
        {'local_scheme': lambda version: ''}
        if getenv('TESTPYPI_UPLOAD') == 'true'
        else False
    )


setup(
    name="stare",
    version="0.0.11",
    use_scm_version=_is_test_pypi(),
    package_dir={"": "src"},
    packages=find_packages(where="src", exclude=["tests"]),
    include_package_data=True,
    description="python sdk for Glance API",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://giordonstark.com",
    author="Giordon Stark",
    author_email="gstark@cern.ch",
    license="BSD 3-clause",
    install_requires=[
        'requests',  # for all HTTP calls to the API
        'cachecontrol[filecache]',  # for caching HTTP requests according to spec to local file
        'click>=6.0',  # for console scripts,
        'python-jose',  # for id token decoding
        'attrs',  # for model inflation/deflation
        'python-dotenv',  # for loading env variables
    ],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    extras_require=extras_require,
    entry_points={'console_scripts': ['stare=stare.commandline:stare']},
    dependency_links=[],
)
